// Milkshape.cpp: implementation of the CMilkshape class.
//
//////////////////////////////////////////////////////////////////////
#include "StdAfx.h"
#include "Milkshape.h"
#include <fstream.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
// byte-align structures
#ifdef _MSC_VER
#	pragma pack( push, packing )
#	pragma pack( 1 )
#	define PACK_STRUCT
#elif defined( __GNUC__ )
#	define PACK_STRUCT	__attribute__((packed))
#else
#	error you must byte-align these structures with the appropriate compiler directives
#endif

typedef unsigned char byte;
typedef unsigned short word;

// File header
struct MS3DHeader
{
	char m_ID[10];
	int m_version;
} PACK_STRUCT;

// BoneVertex information
struct MS3DVertex
{
	byte m_flags;
	float m_vertex[3];
	char BoneID;
	byte m_refCount;
} PACK_STRUCT;

// Triangle information
struct MS3DTriangle
{
	word m_flags;
	word m_vertexIndices[3];
	float m_vertexNormals[3][3];
	float m_s[3], m_t[3];
	byte m_smoothingGroup;
	byte m_groupIndex;
} PACK_STRUCT;

// Material information
struct MS3DMaterial
{
    char m_name[32];
    float Ambient[4];
    float Diffuse[4];
    float Specular[4];
    float Emissive[4];
    float Shininess;	// 0.0f - 128.0f
    float m_transparency;	// 0.0f - 1.0f
    byte m_mode;	// 0, 1, 2 is unused now
    char m_texture[128];
    char m_alphamap[128];
} PACK_STRUCT;

//	Joint information
struct MS3DJoint
{
	byte m_flags;
	char m_name[32];
	char m_parentName[32];
	float m_rotation[3];
	float m_translation[3];
	word m_numRotationKeyframes;
	word m_numTranslationKeyframes;
} PACK_STRUCT;

// Keyframe data
struct MS3DKeyframe
{
	float m_time;
	float m_parameter[3];
} PACK_STRUCT;

// Default alignment
#ifdef _MSC_VER
#	pragma pack( pop, packing )
#endif

#undef PACK_STRUCT


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CMilkshape::CMilkshape()
{

}

void CMilkshape::Destroy_Milkshape()
{
	Destroy_Model();
}
// ----------------------------------------------------------------
// Loads model saved in binary format
// ----------------------------------------------------------------
bool CMilkshape::Load_Model( char *filename, bool bAnimated )
{
	ifstream inputFile( filename, ios::in | ios::binary | ios::nocreate );
	if ( inputFile.fail())
		return false;	// "Couldn't open the model file."

	char pathTemp[255+1];
	int pathLength;
	for ( pathLength = strlen( filename ); --pathLength; ) {
		if ( filename[pathLength] == '/' || filename[pathLength] == '\\' ) {
			break;
		}
	}
	strncpy( pathTemp, filename, pathLength );
	
	int i;
	if ( pathLength > 0 ) {
		pathTemp[pathLength++] = '/';
	}

	strcpy(m_strFilename, filename);
	inputFile.seekg( 0, ios::end );
	long fileSize = inputFile.tellg();
	inputFile.seekg( 0, ios::beg );

	byte *pBuffer = new byte[fileSize];
	inputFile.read( pBuffer, fileSize );
	inputFile.close();

	const byte *pPtr = pBuffer;
	MS3DHeader *pHeader = ( MS3DHeader* )pPtr;
	pPtr += sizeof( MS3DHeader );

	if ( strncmp( pHeader->m_ID, "MS3D000000", 10 ) != 0 )
		return false; // "Not a valid Milkshape3D model file."

	if ( pHeader->m_version < 3 || pHeader->m_version > 4 )
		return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );

	int nVertices = *( word* )pPtr; 
	m_iNumVertices = nVertices;
	m_pVertices = new tBoneVertex[nVertices];
	pPtr += sizeof( word );

	for ( i = 0; i < nVertices; i++ )
	{
		MS3DVertex *pVertex = ( MS3DVertex* )pPtr;
		m_pVertices[i].m_boneID = pVertex->BoneID;
		memcpy( m_pVertices[i].m_Pos, pVertex->m_vertex, sizeof( float )*3 );
		pPtr += sizeof( MS3DVertex );
	}

	int nTriangles = *( word* )pPtr;
	m_iNumTriangles = nTriangles;
	m_pTriangles = new tTriangle[nTriangles];
	pPtr += sizeof( word );

	for ( i = 0; i < nTriangles; i++ )
	{
		MS3DTriangle *pTriangle = ( MS3DTriangle* )pPtr;
		float vertexIndices[3] = { pTriangle->m_vertexIndices[0], pTriangle->m_vertexIndices[1], pTriangle->m_vertexIndices[2] };
		float t[3] = { 1.0f-pTriangle->m_t[0], 1.0f-pTriangle->m_t[1], 1.0f-pTriangle->m_t[2] };
		memcpy( m_pTriangles[i].m_vertexNormals, pTriangle->m_vertexNormals, sizeof( float )*3*3 );
		memcpy( m_pTriangles[i].m_s, pTriangle->m_s, sizeof( float )*3 );
		memcpy( m_pTriangles[i].m_t, t, sizeof( float )*3 );
		memcpy( m_pTriangles[i].m_vertexIndices, vertexIndices, sizeof( int )*3 );
		pPtr += sizeof( MS3DTriangle );
	}

	int nGroups = *( word* )pPtr;
	m_iNumMeshes = nGroups;
	m_pMeshes = new tMesh[nGroups];
	pPtr += sizeof( word );
	for ( i = 0; i < nGroups; i++ )
	{
		pPtr += sizeof( byte );	// flags
		pPtr += 32;				// name

		word nTriangles = *( word* )pPtr;
		pPtr += sizeof( word );
		int *pTriangleIndices = new int[nTriangles];
		for ( int j = 0; j < nTriangles; j++ )
		{
			pTriangleIndices[j] = *( word* )pPtr;
			pPtr += sizeof( word );
		}

		char materialIndex = *( char* )pPtr;
		pPtr += sizeof( char );
	
		m_pMeshes[i].m_materialIndex = materialIndex;
		m_pMeshes[i].m_numTriangles = nTriangles;
		m_pMeshes[i].m_pTriangleIndices = pTriangleIndices;
	}

	int nMaterials = *( word* )pPtr;
	m_iNumMaterials = nMaterials;
	m_pMaterials = new tMaterial[nMaterials];
	pPtr += sizeof( word );
	for ( i = 0; i < nMaterials; i++ )
	{
		MS3DMaterial *pMaterial = ( MS3DMaterial* )pPtr;
		m_pMaterials[i].m_Ambient = RGBA(pMaterial->Ambient[0], pMaterial->Ambient[1], pMaterial->Ambient[2], pMaterial->Ambient[3]);
		m_pMaterials[i].m_Diffuse = RGBA(pMaterial->Diffuse[0], pMaterial->Diffuse[1], pMaterial->Diffuse[2], pMaterial->Diffuse[3]);
		m_pMaterials[i].m_Specular = RGBA(pMaterial->Specular[0], pMaterial->Specular[1], pMaterial->Specular[2], pMaterial->Specular[3]);
		m_pMaterials[i].m_Emissive = RGBA(pMaterial->Emissive[0], pMaterial->Emissive[1], pMaterial->Emissive[2], pMaterial->Emissive[3]);
		m_pMaterials[i].m_fShininess = pMaterial->Shininess;
		
		if(strlen(pMaterial->m_texture) > 0)
		{
			strcpy( pathTemp + pathLength, pMaterial->m_texture + 2 );
			strncpy( m_pMaterials[i].m_strFilename, pathTemp, 255 );

			strncpy(pMaterial->m_texture, m_pMaterials[i].m_strFilename, 255);
			LoadTexture(pMaterial->m_texture, &m_pMaterials[i], true);  
		}
		

		pPtr += sizeof( MS3DMaterial );
	}

	bool haveValidTex = false;
	for (i=0; i<nMaterials; i++)
	{
		if(glIsTexture(m_pMaterials[i].m_id))
		{
			m_eMaterialRender = MAT_TEX;
			haveValidTex = true;
			break;
		}
	}
	if(!haveValidTex)
		m_eMaterialRender = MAT;

	//((CModel *)this)->calculateNormals();

	// { NEW }
	// Load Skeletal Animation Stuff
	if(!bAnimated)
	{
		SAFEDEL(pBuffer);
		return true;
	}

	float animFPS = *( float* )pPtr;
	pPtr += sizeof( float );

	// skip currentTime
	pPtr += sizeof( float );

	int totalFrames = *( int* )pPtr;
	pPtr += sizeof( int );

	m_fTotalTime = totalFrames*1000.0/animFPS;

	m_iNumJoints = *( word* )pPtr;
	pPtr += sizeof( word );

	if(m_iNumJoints > 0)
	{
		m_pJoints = new tJoint[m_iNumJoints];

		struct JointNameListRec
		{
			int m_jointIndex;
			const char *m_pName;
		};

		const byte *pTempPtr = pPtr;

		JointNameListRec *pNameList = new JointNameListRec[m_iNumJoints];
		for ( i = 0; i < m_iNumJoints; i++ )
		{
			MS3DJoint *pJoint = ( MS3DJoint* )pTempPtr;
			pTempPtr += sizeof( MS3DJoint );
			pTempPtr += sizeof( MS3DKeyframe )*( pJoint->m_numRotationKeyframes+pJoint->m_numTranslationKeyframes );

			pNameList[i].m_jointIndex = i;
			pNameList[i].m_pName = pJoint->m_name;
		}

		for ( i = 0; i < m_iNumJoints; i++ )
		{
			MS3DJoint *pJoint = ( MS3DJoint* )pPtr;
			pPtr += sizeof( MS3DJoint );

			int j, parentIndex = -1;
			if ( strlen( pJoint->m_parentName ) > 0 )
			{
				for ( j = 0; j < m_iNumJoints; j++ )
				{
					if ( stricmp( pNameList[j].m_pName, pJoint->m_parentName ) == 0 )
					{
						parentIndex = pNameList[j].m_jointIndex;
						break;
					}
				}
				if ( parentIndex == -1 ) {
					cerr << "Unable to find parent bone in MS3D file" << endl;
					return false;
				}
			}

			m_pJoints[i].m_localRotation = pJoint->m_rotation;
			m_pJoints[i].m_localTranslation = pJoint->m_translation;
			m_pJoints[i].m_parent = parentIndex;
			m_pJoints[i].m_numRotationKeyframes = pJoint->m_numRotationKeyframes;
			m_pJoints[i].m_pRotationKeyframes = new tKeyframe[pJoint->m_numRotationKeyframes];
			m_pJoints[i].m_numTranslationKeyframes = pJoint->m_numTranslationKeyframes;
			m_pJoints[i].m_pTranslationKeyframes = new tKeyframe[pJoint->m_numTranslationKeyframes];

			for ( j = 0; j < pJoint->m_numRotationKeyframes; j++ )
			{
				MS3DKeyframe *pKeyframe = ( MS3DKeyframe* )pPtr;
				pPtr += sizeof( MS3DKeyframe );

				SetJointKeyframe( i, j, pKeyframe->m_time*1000.0f, pKeyframe->m_parameter, true );
			}

			for ( j = 0; j < pJoint->m_numTranslationKeyframes; j++ )
			{
				MS3DKeyframe *pKeyframe = ( MS3DKeyframe* )pPtr;
				pPtr += sizeof( MS3DKeyframe );

				SetJointKeyframe( i, j, pKeyframe->m_time*1000.0f, pKeyframe->m_parameter, false );
			}
		}
		delete[] pNameList;

		SetupJoints();
	//	calculateNormals();
		
		Restart();
	}
	delete[] pBuffer;
	return true;
}


